import json
from dataclasses import asdict, dataclass
from typing import List, Optional

import numpy as np
from sklearn.preprocessing import MultiLabelBinarizer

Case = dict
Cases = List[dict]
Mlb = MultiLabelBinarizer
Tag = str
Label = List[Tag]
Labels = List[Label]

TAG = [
    "acute leukemia",
    "acute lymphoblastic leukemia",
    "acute myeloid leukemia",
    "acute promyelocytic leukemia",
    "basophilia",
    "chronic myeloid leukemia",
    "eosinophilia",
    "erythroid hyperplasia",
    "granulocytic hyperplasia",
    "hemophagocytosis",
    "hypercellular",
    "hypocellular",
    "inadequate",
    "iron deficiency",
    "lymphoproliferative disorder",
    "mastocytosis",
    "metastatic",
    "myelodysplastic syndrome",
    "myeloproliferative neoplasm",
    "normal",
    "plasma cell neoplasm",
]


def get_thresh():
    try:
        with open("thresh.json") as target:
            print("Use thresh.json")
            THRESH = json.load(target)
    except FileNotFoundError:
        print("Use 0.5 Threshold")
    THRESH = 0.5
    return THRESH


@dataclass
class LabelledCase:
    text: Case
    tag: Label

    def serialize(self):
        return {"text": self.text, "tag": "; ".join(self.tag)}


@dataclass
class Mask:
    field: str
    start: int
    end: int
    word: str

    def __call__(self, case: Case):
        case_copy = case.copy()
        field = case_copy[self.field]
        case_copy[self.field] = field[: self.start] + field[self.end :]
        return case_copy

    def mark(self, case: Case, mark):
        case_copy = case.copy()
        field = case_copy[self.field]
        case_copy[self.field] = field[: self.start] + mark + field[self.start :]
        return case_copy


@dataclass
class MaskedParent:
    masks: List[Mask]
    text: Case

    def masked_cases(self):
        return [mask(self.text) for mask in self.masks]

    def mask_words(self) -> List[str]:
        return [mask.word for mask in self.masks]

    def mask_words_field(self) -> List[tuple]:
        return [(mask.word, mask.field) for mask in self.masks]


@dataclass
class MaskRet:
    tag: str
    importance: list


@dataclass
class Trace:
    origin_output: np.array
    masked_outputs: np.array
    masks: np.array
    important_change: Optional[np.array] = None


@dataclass
class States:
    data: np.array
    tag: list
    index: list
    tag_n: list
    from_: list
    pred_tag: list


DATAFILE = {
    "x_dict": "x_dict.json",
    "y_tags": "y_tags.json",
    "x_train_dict": "x_train_dict.json",
    "y_train_tags": "y_train_tags.json",
    "x_test_dict": "x_test_dict.json",
    "y_test_tags": "y_test_tags.json",
}


@dataclass
class RawData:
    x_dict: Cases
    y_tags: Labels
    x_train_dict: Cases
    y_train_tags: Labels
    x_test_dict: Cases
    y_test_tags: Labels

    def __iter__(self):
        return iter(asdict(self).items())

    def show(self, from_: str, idx: int):

        print(json.dumps(self.retrive(from_, idx), indent=2))

    def retrive(self, from_: str, idx: int):
        if from_ == "train":
            x = self.x_train_dict[idx]
            y = self.y_train_tags[idx]
        else:
            x = self.x_test_dict[idx]
            y = self.y_test_tags[idx]
        return {"text": x, "tag": y}

    def to_labelled_cases(self):
        return [LabelledCase(text, tag) for text, tag in zip(self.x_dict, self.y_tags)]


@dataclass
class Params:
    raw_data: RawData
    max_len: int
    upsampling: int
    dropout_prob: float
    identifier: str
    keep_key: bool
    epoch: int
    mlb: Optional[MultiLabelBinarizer] = None



from collections import Counter, defaultdict

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.figure_factory as ff
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from plotly.validators.scatter.marker import SymbolValidator
from sklearn.preprocessing import normalize

TEMPLATE = "plotly_white"


def plot_tag_stat(tag_df: pd.DataFrame):
    s = tag_df.groupby("tag")["count"].sum()
    sort_idx = []
    for i in s.sort_values(ascending=False).index:
        sort_idx.append(i + "train")
        sort_idx.append(i + "test")
    tag_df["idx"] = tag_df["tag"] + tag_df["for"]
    tag_df = tag_df.set_index("idx")
    sort_idx = [idx for idx in sort_idx if idx in tag_df.index]
    tag_df = tag_df.loc[sort_idx]
    fig = px.bar(
        tag_df,
        x="tag",
        y="count",
        color="for",
        title="Label Summary",
        text="count",
        color_discrete_sequence=px.colors.qualitative.Set2,
    )
    fig.update_layout(
        template=TEMPLATE,
        font_family="Arial",
        width=1280,
        height=600,
    )
    return fig


def state_plot(df: pd.DataFrame, thresh=15):

    markers = df.tag.value_counts()
    tag_value_counts = markers >= thresh
    show_legends = set(tag_value_counts.index.to_numpy()[tag_value_counts])
    df["show_legends"] = df["tag"].map(lambda x: x in show_legends)

    raw_symbols = SymbolValidator().values
    colors = px.colors.qualitative.Plotly
    symbol_dict = {}
    color_dict = {}
    color_len = len(colors)
    for idx, tag in enumerate(markers.index):
        symbol_idx = idx // color_len
        color_idx = idx % color_len
        symbol_dict[tag] = raw_symbols[symbol_idx]
        color_dict[tag] = colors[color_idx]
    df["color"] = df.tag.map(color_dict)
    df["symbol"] = df.tag.map(symbol_dict)

    fig = go.Figure()
    sel_tags = sorted(show_legends, key=len)
    for sel_tag in sel_tags:
        tmp_df = df.loc[df.tag == sel_tag, :]
        fig.add_trace(
            go.Scatter(
                x=tmp_df["D1"],
                y=tmp_df["D2"],
                mode="markers",
                marker_color=tmp_df["color"],
                marker_symbol=tmp_df["symbol"],
                hovertemplate="<b>tag:%{text}</b><br>pred_tag:%{customdata}",
                text=tmp_df["tag"].to_list(),
                customdata=tmp_df["pred_tag"].to_list(),
                showlegend=True,
                name=sel_tag,
            )
        )

    no_legend_df = df.loc[~df["show_legends"], :]
    fig.add_trace(
        go.Scatter(
            x=no_legend_df["D1"],
            y=no_legend_df["D2"],
            mode="markers",
            opacity=0.5,
            marker_color=no_legend_df["color"],
            marker_symbol=no_legend_df["symbol"],
            showlegend=False,
            hovertemplate="<b>tag:%{text}</b><br>pred_tag:%{customdata}",
            text=no_legend_df["tag"].to_list(),
            customdata=no_legend_df["pred_tag"].to_list(),
            name="",
        )
    )

    fig.update_xaxes(showticklabels=False)
    fig.update_yaxes(showticklabels=False)
    fig.update_layout(
        template=TEMPLATE,
        font_family="Arial",
        legend=dict(
            orientation="h",
        ),
        width=1280,
        height=600,
    )
    return fig


def plot_judge_num(j_tag_num, mode="Correct"):
    data = defaultdict(list)
    for k, v in j_tag_num.items():
        counter = Counter(v)
        data["Correct Num"].extend(counter.keys())
        data["Count"].extend(counter.values())
        data["Tag Num"].extend(k for _ in range(len(counter)))

    tmp_df = (
        pd.pivot_table(
            pd.DataFrame(data), index=["Tag Num", "Correct Num"], values="Count"
        )
        .unstack()
        .fillna(0)
    )
    tmp_df.columns = [f"{c[1]} {mode}" for c in tmp_df.columns]

    judge_df_rate = tmp_df.copy()
    j_df_sum = judge_df_rate.sum(axis=1)

    for idx, row in enumerate(judge_df_rate.iterrows()):
        judge_df_rate.values[idx] = row[1] / j_df_sum.iloc[idx]

    data = []
    judge_df = tmp_df.reset_index()
    tag_num = judge_df["Tag Num"]
    for col in judge_df.columns[1:]:
        y = judge_df[col]
        text = [f"{v * 100:.02f} %" for v in judge_df_rate[col]]
        data.append(
            go.Bar(
                name=col,
                x=tag_num,
                y=y,
                text=text,
                textposition="auto",
            ),
        )
    fig = go.Figure(data=data)
    fig.update_layout(
        template=TEMPLATE,
        font_family="Arial",
        barmode="stack",
        xaxis_title="Tag Number",
        yaxis_title="Count",
        legend_title=f"{mode} Predition",
    )
    return fig


def kw_plot(top_key):
    top_n = 5
    labels = []
    values = []
    parents = []
    for idx, (k, v) in enumerate(top_key.items()):
        labels.extend(i[0] for i in v[:top_n])
        tmp_v = [i[1] for i in v]
        tmp_v = normalize([tmp_v])[0]
        values.extend(round(v, 3) for v in tmp_v[:top_n])
        parents.extend(k for _ in range(top_n))
    kws_df = pd.DataFrame({"influence": values, "reason": labels, "tag": parents})
    heat_df = kws_df.groupby("tag").agg(list)

    vs = heat_df["influence"].values
    mat = np.zeros(
        (len(vs), len(vs[0])),
    )
    for i, r in enumerate(vs):
        for j, v in enumerate(r):
            mat[i][j] = v

    sort_ = mat[:, 0].argsort()
    z = heat_df["influence"].values[sort_].tolist()
    y = heat_df.index[sort_].to_list()
    z_text = heat_df["reason"].values[sort_].tolist()
    fig = ff.create_annotated_heatmap(
        z,
        y=y,
        x=list(range(1, 1 + top_n, 1)),
        annotation_text=z_text,
        colorscale="Blues",
        showscale=True,
    )
    fig.update_layout(
        template=TEMPLATE,
        font_family="Arial",
        width=1280,
        height=600,
    )
    return fig


def plot_summary(data):
    tree_data = defaultdict(dict)
    for n in range(1, 7):
        labels = []
        parents = []
        values = []
        for type_ in ("less", "more", "equal"):
            if type_ != "equal":
                count = 0
                for k_, v_ in data[type_][n].items():
                    tmp_v = []
                    for k, v in v_.items():
                        labels.append(f"{k} correct")
                        tmp_v.append(v)
                        parents.append(f"pred {k_ + n}")

                    labels.append(f"pred {k_ + n}")
                    sum_v = sum(tmp_v)
                    values.extend(tmp_v)
                    values.append(sum_v)
                    count += sum_v
                    parents.append(type_)
                labels.append(type_)
                values.append(count)
                parents.append("")
            else:
                tmp_v = []
                for k, v in data[type_][n].items():
                    labels.append(f"{k} correct")
                    tmp_v.append(v)
                    parents.append(type_)

                values.extend(tmp_v)

                values.append(sum(tmp_v))
                labels.append(type_)
                parents.append("")

        tree_data[n]["labels"] = labels
        tree_data[n]["parents"] = parents
        tree_data[n]["values"] = values

    col_num = 3
    row_num = int(6 / col_num)
    fig = make_subplots(
        cols=col_num,
        rows=row_num,
        horizontal_spacing=0.02,
        vertical_spacing=0.05,
        specs=[[{"type": "domain"} for _ in range(col_num)] for _ in range(row_num)],
        subplot_titles=[f"{tag_num} Tag" for tag_num in tree_data.keys()],
    )

    for idx, tmp_tree in enumerate(tree_data.values()):

        col = idx % col_num + 1
        row = idx // col_num + 1
        fig.add_trace(
            go.Sunburst(
                labels=tmp_tree["labels"],
                parents=tmp_tree["parents"],
                values=tmp_tree["values"],
                branchvalues="total",
            ),
            row=row,
            col=col,
        )

    fig.update_layout(
        template=TEMPLATE,
        font_family="Arial",
        width=1280,
        height=600,
    )
    return fig


def plot_tag_performance(performance: pd.DataFrame, overall, auc=False):
    y_title = "AUC" if auc else "F1 Score"
    performance.sort_values(
        y_title,
        inplace=True,
    )
    fig = go.Figure()
    x = performance["Tag"]
    marker_symbols = ["square", "cross"]
    perf_average = overall[y_title]
    fig.add_shape(
        type="line",
        x0=0.01,
        y0=perf_average,
        x1=0.99,
        y1=perf_average,
        xref="paper",
        line=dict(color="lightgray", width=2, dash="dash"),
    )
    for idx, measure in enumerate(["Precision", "Recall"]):
        fig.add_trace(
            go.Scatter(
                x=x,
                y=performance[measure],
                marker_color="lightgray",
                marker_symbol=marker_symbols[idx],
                marker_size=10,
                mode="markers",
                name=measure,
            )
        )

    fig.add_trace(
        go.Scatter(
            x=x,
            y=performance[y_title],
            marker_color="crimson",
            mode="markers+text",
            text=[f"{v:.02f}" for v in performance[y_title]],
            marker_size=10,
            name=y_title,
        )
    )
    fig.update_traces(textposition="middle right")

    x_loc = len(performance) - 2
    fig.update_layout(
        template=TEMPLATE,
        font_family="Arial",
        width=1280,
        height=600,
        xaxis_title="Semantic Label",
        yaxis_title="Metrics",
        showlegend=True,
        annotations=[
            dict(
                x=x_loc,
                y=perf_average,
                xref="x",
                yref="y",
                text=f"Micro {y_title}: {perf_average:.03f}",
                showarrow=False,
                font=dict(size=15),
            ),
        ],
    )
    return fig


def plot_num_performance(performance_n: pd.DataFrame):
    fig = go.Figure()
    x = performance_n.index
    size = performance_n["Count"]
    for c in ["F1 Score"]:
        y = performance_n[c]
        fig.add_trace(
            go.Scatter(
                x=x,
                y=y,
                name=c,
                mode="markers+text",
                marker=dict(
                    size=size,
                    sizemode="area",
                    sizeref=2.0 * max(size) / (40.0 ** 2),
                    sizemin=5,
                ),
                text=[f"{v:.03f}" for v in y],
                textposition="middle right",
            )
        )

    fig.update_layout(
        template=TEMPLATE,
        font_family="Arial",
        width=1280,
        height=600,
        uniformtext_minsize=11,
        uniformtext_mode="show",
    )
    fig.update_xaxes(title_text="Tag Number")

    return fig



import json
import pickle
import time
from pathlib import Path
from typing import List
from zipfile import ZipFile

import requests
from tqdm.autonotebook import tqdm

from .domain import DATAFILE, LabelledCase, RawData


def prepare_model(out_path="model"):
    url = "https://storage.googleapis.com/pathopatho/model.zip"
    out_f = "model.zip"
    download_file(url, out_f)
    unzip_model(out_f, out_path)


def download_file(url, out_f):
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(out_f, "wb") as f:
            for chunk in tqdm(r.iter_content(chunk_size=8192)):
                f.write(chunk)
    return out_f


def unzip_model(p, out):
    with ZipFile(p, "r") as zipfile:
        zipfile.extractall(out)
    m = Path(out)
    for f in m.glob("*/*"):
        f.rename(m / f.name)


def prepare_dataset(out_f="dataset.zip"):
    url = "https://storage.googleapis.com/pathopatho/dataset.zip"
    download_file(url, out_f)


def prepare_unlabelled(out_f="data/unlabelled.json"):
    url = "https://storage.googleapis.com/pathopatho/unlabelled.json"
    download_file(url, out_f)


def prepare_state(out_f="data/unstate.pkl"):
    url = "https://storage.googleapis.com/pathopatho/unstate.pkl"
    download_file(url, out_f)


def get_timestamp():
    return time.strftime("%Y%m%d-%H%M%S")


def load_json(path):
    with open(path, "r") as js_:
        return json.load(js_)


def dump_json(path, obj):
    with open(path, "w") as js_:
        json.dump(obj, js_, indent=4)


def load_datazip(datazip_path: str, datafile: dict = DATAFILE):
    with ZipFile(datazip_path, "r") as datazip:
        tmp = []
        for f_name in datafile.values():
            with datazip.open(f_name, "r") as file:
                tmp.append(json.loads(file.read().decode("utf-8")))
        return RawData(*tmp)


def dump_datazip(rawdata: RawData, zip_name=None):
    if zip_name is None:
        zip_name = f"dataset{get_timestamp()}.zip"
    with ZipFile(zip_name, "w") as datazip:
        for name, data in rawdata:

            with datazip.open(f"{name}.json", "w") as file:
                file.write(json.dumps(data).encode("utf-8"))
    return zip_name


def dump_labelled_cases(labelled_cases: List[LabelledCase], path: str):
    obj = list(map(LabelledCase.serialize, labelled_cases))
    dump_json(path, obj)


def load_state(state_p: str):
    with open(state_p, "rb") as plk:
        state = pickle.load(plk)
    return state


def dump_state(states, state_p="state"):
    with open(state_p, "wb") as plk:
        pickle.dump(states, plk)
    return state_p


def build_eval_json(sampled_cases, pred_prob, pred_out):
    eval_json = []
    for idx, (case, prob, pred) in enumerate(zip(sampled_cases, pred_prob, pred_out)):
        eval_json.append(
            {
                "id": idx,
                "text": case,
                "prob": [(n, str(p)) for n, p in prob],
                "pred": [bool(v) for v in pred],
            }
        )
    return eval_json



"""Data Transformation and Utility"""
import random
from collections import Counter, defaultdict
from itertools import chain
from typing import Dict, List

import numpy as np
import pandas as pd
from sklearn.preprocessing import MultiLabelBinarizer

from .domain import Cases, Label, LabelledCase, Labels, RawData
from .io_utils import dump_datazip, get_timestamp, load_json

random.seed(42)


def compose(case: dict, keep_key=True, shuffle=False) -> str:
    if keep_key:
        tmp = [f"{k}: {v}" for k, v in case.items()]
    else:
        tmp = list(case.values())
    if shuffle:
        random.shuffle(tmp)
    return " ".join(tmp)


def upsampling(x: list, y: list, keep_key=True, target=100):
    new_x: List[str] = []
    new_y = []
    group_by_idx = grouping_idx(y)
    if target < 0:
        shuffle = False
        data_size = len(group_by_idx) * abs(target)
        while len(new_x) < data_size:
            new_x.extend(
                map(lambda case: compose(case, keep_key=keep_key, shuffle=shuffle), x)
            )
            new_y.extend(y)
        new_x = new_x[:data_size]
        new_y = new_y[:data_size]
    else:
        shuffle = True
        for group_idx in group_by_idx.values():
            upsample_idx = random.choices(group_idx, k=target)
            new_x.extend(
                map(
                    lambda idx: compose(x[idx], keep_key=keep_key, shuffle=shuffle),
                    upsample_idx,
                )
            )
            new_y.extend(map(lambda idx: y[idx], upsample_idx))
    return new_x, new_y


def grouping_idx(y) -> Dict[str, list]:
    group_by = defaultdict(list)
    for idx, tags in enumerate(y):
        for tag in tags:
            group_by[tag].append(idx)
    return group_by


def count_tags(tags: Labels):
    return Counter(chain(*tags))


def count_token_len(cases: Cases):
    lens = list(map(lambda case: len(" ".join(case.values()).split(" ")), cases))
    return Counter(lens)


def unwrap_labelled_cases(labelled_cases: List[LabelledCase]):
    cases = [lc.text for lc in labelled_cases]
    tags = [lc.tag for lc in labelled_cases]
    return cases, tags


def label_to_tags(label: str):
    return refine_tag(label.split(";"))


def refine_tag(label: Label):
    return [
        tag
        for tag in add_acute_LL(
            list(
                map(
                    lambda x: tag_patch(edit_tag_str(x)),
                    label,
                )
            )
        )
        if tag != ""
    ]


def edit_tag_str(tag: str):
    typos = dict(
        [
            ("inadquate", "inadequate"),
            ("eosinophila", "eosinophilia"),
            ("hemophagoctyosis", "hemophagocytosis"),
            ("hypercellula", "hypercellular"),
            ("hypocellularr", "hypocellular"),
            ("lymphoma", "acute lymphoblastic leukemia"),
        ]
    )

    tag = tag.lower().strip().replace(".", "").replace("syndrome no", "syndrome")

    tag = typos.get(tag, tag)
    return tag


def tag_patch(tag: str):
    if tag == "plasma" or tag == "plasma cell disorder":
        return "plasma cell neoplasm"
    return tag


def add_acute_LL(tags: List[str]):
    """the tag \"lymphoproliferative disorder\"
    should be added to any case tagged as \"acute lymphoblastic leukemia\".

    Returns:
        [type]: [description]
    """
    w1 = "acute lymphoblastic leukemia"
    w2 = "lymphoproliferative disorder"
    if w1 in tags and w2 not in tags:
        return tags + [w2]
    return tags


def cases_minus(minuend: Cases, subtrahend: Cases):
    def not_same_content(case: dict) -> bool:
        return "".join(case.values()) not in used_cases

    used_cases = {"".join(case.values()) for case in subtrahend}
    assert "" not in used_cases, "has empty case in other cases"
    difference = list(filter(not_same_content, minuend))

    return difference


def labelled_cases_to_xy(labelled_cases: List[LabelledCase]):
    x = []
    y = []
    for labelled_case in labelled_cases:
        x.append(labelled_case.text)
        y.append(labelled_case.tag)
    return x, y


def xy_to_labelled_cases(x, y) -> List[LabelledCase]:
    return [LabelledCase(text, tag) for text, tag in zip(x, y)]


def split_and_dump_dataset(x, y, test_size=0.2, train_first=True, output=None) -> str:
    x_train_dict, x_test_dict, y_train_tags, y_test_tags = train_test_split(
        x, y, test_size=test_size, train_first=train_first
    )
    rd = RawData(x, y, x_train_dict, y_train_tags, x_test_dict, y_test_tags)
    if output is None:
        output = f"dataset{get_timestamp()}.zip"
    dump_datazip(rd, output)
    return output


def load_labelled_cases(path):
    records = load_json(path)
    labelled_cases = []
    for record in records:
        labelled_cases.append(
            LabelledCase(record["text"], label_to_tags(record["tag"]))
        )
    return labelled_cases


def train_test_split(x, y, test_size=0.2, train_first=True):
    if train_first:
        train_idx, test_idx = _train_first_split(y, test_size)
    else:
        train_idx, test_idx = _test_first_split(y, test_size)
    return (
        [x[idx] for idx in train_idx],
        [x[idx] for idx in test_idx],
        [y[idx] for idx in train_idx],
        [y[idx] for idx in test_idx],
    )


def _train_first_split(y, test_size):
    data_size = len(y)
    tag_stat = count_tags(y)
    keep_tags = get_rare_tags(tag_stat)
    train_idx = []
    left_idx = []
    # active learning period -> train_first
    indices = list(range(data_size))
    random.shuffle(indices)
    for idx in indices:
        tags = y[idx]
        if any(tag in keep_tags for tag in tags):
            train_idx.append(idx)
        else:
            left_idx.append(idx)
    random.shuffle(left_idx)
    bin_point = round(test_size * data_size)
    test_idx = left_idx[:bin_point]
    train_idx.extend(left_idx[bin_point:])
    return train_idx, test_idx


def _test_first_split(y, test_size):
    # evaluation learning period -> test_first
    data_size = len(y)
    tag_stat = count_tags(y)
    keep_tag_dict = {tag: 0 for tag in tag_stat.keys()}
    test_idx = []
    left_idx = []
    min_test_case = round(min(tag_stat.values()) * test_size)
    print(f"min_test_case: {min_test_case}")
    indices = list(range(data_size))
    random.shuffle(indices)
    for idx in indices:
        tags = y[idx]
        if any(tag in keep_tag_dict for tag in tags):
            test_idx.append(idx)
            for tag in tags:
                # make sure at least test_size cases are in test for each labels
                try:
                    keep_tag_dict[tag] += 1
                    if keep_tag_dict[tag] >= min_test_case:
                        del keep_tag_dict[tag]
                except KeyError:
                    pass
        else:
            left_idx.append(idx)
    random.shuffle(left_idx)
    bin_point = round((1 - test_size) * data_size)
    train_idx = left_idx[:bin_point]
    test_idx.extend(left_idx[bin_point:])
    return train_idx, test_idx


def get_rare_tags(tag_count: dict, thresh=1):
    return [tag for tag, count in tag_count.items() if count <= thresh]


def show_replica(cases: Cases):
    tmp = Counter("".join(case.values()) for case in cases)
    for k, v in tmp.items():
        if v > 1:
            print(k)


def get_Mlb(rawdata: RawData):
    return MultiLabelBinarizer().fit(rawdata.y_tags)


def topN(preds, classes, n=3):
    tops = np.argsort(preds)
    ret = []
    for i, top in enumerate(tops):
        sel_idx = top[-n:][::-1]
        ret.append(list(zip(classes[sel_idx], preds[i][sel_idx])))
    return ret


def rawdata_stat(rawdata: RawData):
    def counter_to_df(counter: dict):
        df = pd.DataFrame.from_dict(counter, orient="index")
        df.reset_index(inplace=True)
        df.columns = ["tag", "count"]
        return df

    train_counter = count_tags(rawdata.y_train_tags)
    test_counter = count_tags(rawdata.y_test_tags)
    print(f"Training Size: {len(rawdata.y_train_tags)}")
    print(f"Test Size: {len(rawdata.y_test_tags)}")
    print(f"Train Tag Number: {len(train_counter)}")
    print(f"Test Tag Number: {len(test_counter)}")

    train_tag_df = counter_to_df(train_counter)
    train_tag_df["for"] = "train"
    test_tag_df = counter_to_df(test_counter)
    test_tag_df["for"] = "test"

    tag_df = pd.concat([train_tag_df, test_tag_df])
    return tag_df


def refine_rawdata(rawdata: RawData):
    rawdata.y_tags = list(map(refine_tag, rawdata.y_tags))
    rawdata.y_test_tags = list(map(refine_tag, rawdata.y_test_tags))
    rawdata.y_train_tags = list(map(refine_tag, rawdata.y_train_tags))
    return rawdata


def adjust_normal(tags):
    w = "normal"
    for tag in tags:
        if len(tag) > 1 and w in tag:
            tag.remove(w)



import logging
import os

from .web import Server, WebConfig


def main(reset=False):
    logger = logging.getLogger()
    os.makedirs("data", exist_ok=True)
    web_config = WebConfig(
        "dataset.zip", "model", "data/unlabelled.json", "data/unstate.pkl"
    )
    if reset:
        logger.warning("RESET")
        web_config.reset()
    web_config.init_static()

    server = Server(web_config)
    server.plot()
    server.app.run_server(debug=False)


if __name__ == "__main__":
    from fire import Fire

    Fire(main)



import json
import re
from typing import List

import dash_core_components as dcc
import dash_html_components as html
import plotly.express as px

from .domain import Mask


def html_input(id_):
    return html.Div(["Input: ", dcc.Input(id=id_, type="text")])


def html_dots(id_, fig):
    return dcc.Graph(id=id_, figure=fig)


def html_case(id_="click-data"):
    return html.Div(
        id=id_,
        style={
            "white-space": "pre",
            "overflowX": "scroll",
        },
    )


def html_mask(id_="mask"):
    return dcc.Graph(id=id_)


def dict_to_str(case):
    return json.dumps(case, indent=2)


def draw_color(case, key_masks: List[Mask]):
    finder = re.compile(r"[\w<>]+")
    children = []
    pos_mark = "<POS>"
    pos_style = {"color": "red"}
    for mask in key_masks:
        case = mask.mark(case, pos_mark)

    text = dict_to_str(case)
    cur = 0
    for part in finder.finditer(text):
        g = part.group(0)
        start, end = part.span()
        if pos_mark in g:
            children.append(text[cur:start])
            children.append(html.Span(f'{g.replace(pos_mark, "")} ', style=pos_style))
        else:
            children.append(text[cur:end])
        cur = end
    children.append(text[end:])
    return html.P(children)


def html_checkbox(id_):
    # options = [{"label": t, "value": t} for t in TAG]
    return dcc.Checklist(id=id_)


def html_submit(id_):
    return html.Button("Submit", id=id_)


def empty_bar():
    return px.bar(x=["None"], y=[0])



import re
from collections import defaultdict
from typing import DefaultDict, Dict, List

import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from sklearn.preprocessing import MultiLabelBinarizer
from tqdm.autonotebook import tqdm

from .data_utils import count_tags
from .domain import Case, Mask, MaskedParent, MaskRet, Trace, get_thresh
from .model import StandaloneModel, label_output


class MaskMaker:
    def __init__(self, pattern=r"\w+"):
        self.finder = re.compile(pattern)

    def __call__(self, case: Case):
        masks = []
        for field, sent in case.items():
            for match in self.finder.finditer(sent):
                start, end = match.span()
                mask = Mask(field, start, end, match.group(0))
                masks.append(mask)
        masked_parent = MaskedParent(masks, case)
        return masked_parent


class MaskExplainer:
    def __init__(
        self, mlb: MultiLabelBinarizer, mask_maker: MaskMaker = None, thresh=None
    ):
        if mask_maker is None:
            self.mask_maker = MaskMaker()
        else:
            self.mask_maker = mask_maker
        self.mlb = mlb
        self.thresh = get_thresh() if thresh is None else thresh

    def explain(self, model: StandaloneModel, case: Case):
        origin_output = model.predict([case], tqdm_disable=True)
        masked_parent = self.mask_maker(case)
        masked_cases = masked_parent.masked_cases()
        masked_outputs = model.predict(masked_cases, tqdm_disable=True)
        masks = np.array(masked_parent.masks)
        trace = Trace(origin_output, masked_outputs, masks)
        ret = self.analysis_trace(trace)
        return ret

    def analysis_trace(self, trace: Trace) -> List[MaskRet]:
        def sort_col_descend(values, col):
            return np.argsort(values[:, col])[::-1]

        pred = label_output(trace.origin_output, self.thresh)
        bool_idx = pred[0]
        trace.important_change = (
            trace.origin_output[:, bool_idx] - trace.masked_outputs[:, bool_idx]
        )
        pred_tag = self.mlb.inverse_transform(pred)[0]
        rets = []
        for idx, tag in enumerate(pred_tag):
            rank = sort_col_descend(trace.important_change, idx)
            importance = [
                (mask, value)
                for mask, value in zip(
                    trace.masks[rank], trace.important_change[:, idx][rank]
                )
            ]
            rets.append(MaskRet(tag, importance))
        self.trace = trace
        return rets

    def show_trace(self):
        trace = self.trace
        return trace.origin_output, trace.masked_outputs, trace.important_change


def plot_explanation(rets: List[MaskRet], dash=False):
    fig = make_subplots(
        rows=len(rets), cols=1, subplot_titles=tuple(ret.tag for ret in rets)
    )
    for loc, mask_ret in enumerate(rets, start=1):
        importance = mask_ret.importance
        words_p = [p[0].word for p in importance if p[1] >= 0]
        values_p = [p[1] for p in importance if p[1] >= 0]
        words_n = [p[0].word for p in importance if p[1] < 0]
        values_n = [p[1] for p in importance if p[1] < 0]

        fig.add_trace(go.Bar(x=words_p, y=values_p, name="pos"), row=loc, col=1)
        fig.add_trace(go.Bar(x=words_n, y=values_n, name="neg"), row=loc, col=1)
        # fig.update_xaxes(title_text="word", row=loc, col=1)
        fig.update_yaxes(title_text="influence", row=loc, col=1)
    fig.layout.update(showlegend=False)
    if dash:
        return fig
    fig.show()


def top_keywords(
    mask_explainer: MaskExplainer, model: StandaloneModel, cases, top_n=-1
):
    rets = collect_rets(mask_explainer, model, cases)
    keywords = sum_keywords(rets, top_n)
    return refine_top(keywords, top_n)


def collect_rets(mask_explainer: MaskExplainer, model: StandaloneModel, cases):
    rets = []
    for case in tqdm(cases):
        rets.extend(mask_explainer.explain(model, case))
    return rets


def sum_keywords(rets, top_n=5):
    dashboard: Dict[str, DefaultDict[str, float]] = {}

    for ret in rets:
        importance = ret.importance[:top_n]
        if ret.tag in dashboard:
            tmp_dict = dashboard[ret.tag]
        else:
            tmp_dict = defaultdict(float)
            dashboard[ret.tag] = tmp_dict
        for k, v in importance:
            tmp_dict[k.word] += v

    return dashboard


def refine_top(top, top_n=5):
    refine_ret = {}
    for k, v in top.items():
        crop_len = len(v) if top_n == -1 else top_n
        tmp = sorted(v.items(), key=lambda x: x[1], reverse=True)[:crop_len]
        sum_ = sum(v for _, v in tmp)
        refine_ret[k] = [(k, v / sum_) for k, v in tmp]
    return refine_ret


def filter_top(top: dict, tags, thresh=20):
    tag_counter = count_tags(tags)
    large_enough = [k for k, v in tag_counter.items() if v >= thresh]
    left = {}
    for tag, kw in top.items():
        if tag in large_enough:
            left[tag] = kw
    return left



from collections import Counter, defaultdict

import numpy as np
import pandas as pd
import plotly.express as px
from sklearn import metrics
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

from .data_utils import count_tags, rawdata_stat
from .domain import Mlb, RawData, States
from .io_utils import dump_json
from .model import StandaloneModel, label_output
from .visualization import plot_tag_performance, plot_tag_stat


def eval_model(model, ds, repeat, mlb, output_p, marker):
    tag_stat = rawdata_stat(ds)
    tag_stat.to_csv(f"{output_p}/{marker}_{repeat}_data_stat.csv")
    fig = plot_tag_stat(tag_stat)
    fig.update_layout(
        width=1280,
        height=600,
    )
    fig.write_image(f"{output_p}/{marker}_{repeat}_data_stat.pdf")
    performance, metric, pred_tags = judge_on_tag(model, mlb, ds, n=repeat)
    dump_json(f"{output_p}/{marker}_{repeat}_overall.json", metric)
    performance.to_csv(f"{output_p}/{marker}_{repeat}_Perf_tag.csv")
    fig = plot_tag_performance(performance, metric, auc=False)
    fig.write_image(f"{output_p}/{marker}_{repeat}_Perf_tag.pdf")

    _, _, _, df = summary(
        ds.x_test_dict,
        ds.y_test_tags,
        pred_tags,
    )
    df.to_csv(f"{output_p}/{marker}_{repeat}_summary.csv")


def get_unlabelled_state(model: StandaloneModel, cases: list, mlb: Mlb, thresh=None):
    def tags_to_str(tags):
        return ", ".join(sorted(tags))

    k = len(cases)
    pooled_outputs = model.predict(cases, pooled_output=True)
    pred_tags = model.predict_tags(pooled_outputs, mlb, thresh=thresh)
    pred_tag_note = list(map(tags_to_str, pred_tags))
    index = list(range(k))
    tag_y = pred_tag_note
    tag_n = list(map(lambda tags: len(tags), pred_tags))
    from_ = ["unlabelled" for _ in range(k)]
    states = States(pooled_outputs, tag_y, index, tag_n, from_, pred_tag_note)
    return states


def get_tag_states(model: StandaloneModel, rawdata: RawData, mlb: Mlb, thresh=None):
    def tags_to_str(tags):
        return ", ".join(sorted(tags))

    x = rawdata.x_train_dict + rawdata.x_test_dict
    y = rawdata.y_train_tags + rawdata.y_test_tags
    index = list(range(len(rawdata.x_train_dict)))
    index.extend(range(len(rawdata.x_test_dict)))
    from_ = ["train" for _ in range(len(rawdata.x_train_dict))]
    from_.extend("test" for _ in range(len(rawdata.x_test_dict)))
    tag_n = list(map(lambda tags: len(tags), y))
    tag_y = list(map(tags_to_str, y))
    pooled_outputs = model.predict(x, pooled_output=True)
    pred_tags = model.predict_tags(x, mlb, thresh=thresh)
    pred_tag_note = list(map(tags_to_str, pred_tags))
    states = States(pooled_outputs, tag_y, index, tag_n, from_, pred_tag_note)
    return states


def dimension_reduction(states: States, method_n="PCA", n_components=3):
    if method_n.lower() == "tsne":
        method = TSNE
    else:
        method = PCA
    dimension_reducer = method(n_components=n_components)
    result = dimension_reducer.fit_transform(states.data)
    if isinstance(dimension_reducer, PCA):
        print(
            f"Explained variation per principal component: {dimension_reducer.explained_variance_ratio_}"
        )
    df = pd.DataFrame(
        {
            "tag": states.tag,
            "index": states.index,
            "tag_num": states.tag_n,
            "from": states.from_,
            "pred_tag": states.pred_tag,
        }
    )
    for n in range(n_components):
        df[f"D{n+1}"] = result[:, n]
    return df


def dimension_reduction_plot(df, n_components=3):
    if n_components == 3:
        fig = px.scatter_3d(
            df,
            x="D1",
            y="D2",
            z="D3",
            color="tag",
            symbol="tag_num",
            hover_data=["index", "from", "pred_tag"],
        )
    elif n_components == 2:
        fig = px.scatter(
            df,
            x="D1",
            y="D2",
            color="tag",
            symbol="tag_num",
            hover_data=["index", "from", "pred_tag"],
        )
    else:
        print("support only 2 or 3 dimension ploting")
        return
    fig.layout.update(showlegend=False)
    return fig


def judge_on_tag(
    model: StandaloneModel, mlb: Mlb, rawdata: RawData, thresh=None, n=-1, micro=True
):
    x = rawdata.x_test_dict
    y = rawdata.y_test_tags
    total_y = y + rawdata.y_train_tags
    if n == -1:
        pred_prob = model.predict(x)
    else:
        pred_prob = model.over_predict(x, n=n)

    preds = label_output(pred_prob, thresh)
    y_vector = mlb.transform(y)
    pred_tags = mlb.inverse_transform(preds)

    # Fileter
    class_num = len(mlb.classes_)
    sel_tag_idx = [i for i in range(class_num) if len(np.unique(y_vector[:, i])) != 1]
    y_vector = y_vector[:, sel_tag_idx]
    pred_prob = pred_prob[:, sel_tag_idx]
    preds = preds[:, sel_tag_idx]
    sel_class = [mlb.classes_[i] for i in sel_tag_idx]

    # Computation
    if micro:
        average_auc = metrics.roc_auc_score(
            y_vector, pred_prob, average="micro", multi_class="ovr"
        )
    else:
        average_auc = metrics.roc_auc_score(
            y_vector, pred_prob, average="macro", multi_class="ovr"
        )

    aucs = [
        metrics.roc_auc_score(y_vector[:, i], pred_prob[:, i])
        for i in range(len(sel_tag_idx))
    ]

    precision, recall, f1, _ = metrics.precision_recall_fscore_support(
        y_vector, preds, average="micro"
    )
    mcm = metrics.multilabel_confusion_matrix(y_vector, preds)
    ability = list(map(compress, mcm))
    tag_count = count_tags(total_y)
    sample_sizes = [tag_count[class_] for class_ in sel_class]
    # Aggregation
    performance = pd.DataFrame(
        {
            "Tag": sel_class,
            "Precision": [pair[0] for pair in ability],
            "Recall": [pair[1] for pair in ability],
            "F1 Score": [pair[2] for pair in ability],
            "Testing Size": [pair[3] for pair in ability],
            "Sample Size": sample_sizes,
            "TN": [pair[4] for pair in ability],
            "FP": [pair[5] for pair in ability],
            "FN": [pair[6] for pair in ability],
            "TP": [pair[7] for pair in ability],
            "AUC": aucs,
        }
    )
    performance["Training Size"] = (
        performance["Sample Size"] - performance["Testing Size"]
    )

    metric_ret = {
        "precision": precision,
        "recall": recall,
        "F1 Score": f1,
        "AUC": average_auc,
    }
    return performance, metric_ret, pred_tags


def compress(cm):
    tn, fp = cm[0]
    fn, tp = cm[1]
    amount = fn + tp
    precision = safe_divide(tp, tp + fp)
    recall = safe_divide(tp, amount)
    f1 = safe_divide(2 * precision * recall, precision + recall)
    return (precision, recall, f1, amount, tn, fp, fn, tp)


def safe_divide(a, b):
    if b == 0:
        return 0
    return a / b


def summary(cases, true_tags, pred_tags):
    example = []
    judges = []
    less_tag_num = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
    more_tag_num = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
    equal_tag_num = defaultdict(lambda: defaultdict(int))
    data = {
        "less": less_tag_num,
        "more": more_tag_num,
        "equal": equal_tag_num,
    }
    corrects = []
    pred_tag_numbers = []
    tag_numbers = []
    for case, pred_tag, true_tag in zip(cases, pred_tags, true_tags):
        tag_num = len(true_tag)
        pred_tag_num = len(pred_tag)
        corr = sum(tag in true_tag for tag in pred_tag)
        if pred_tag_num < tag_num:
            judge = f"Less: {corr}/{pred_tag_num} correct. Label:{tag_num} tags"
            data["less"][tag_num][pred_tag_num - tag_num][corr] += 1
        elif pred_tag_num > tag_num:
            judge = f"More: {corr}/{pred_tag_num} correct. Label:{tag_num} tags"
            data["more"][tag_num][pred_tag_num - tag_num][corr] += 1
        else:
            if corr == pred_tag_num:
                judge = "correct"
            else:
                judge = f"Equal: {corr}/{pred_tag_num} correct. Label:{tag_num} tags"
            data["equal"][tag_num][corr] += 1

        corrects.append(corr)
        pred_tag_numbers.append(pred_tag_num)
        tag_numbers.append(tag_num)

        example.append((case, "; ".join(pred_tag), "; ".join(true_tag), judge))
        judges.append(judge)

    df = pd.DataFrame(
        {
            "Correct Count": corrects,
            "Pred Tag Number": pred_tag_numbers,
            "Tag Number": tag_numbers,
        }
    )
    return (example, Counter(judges), data, df)


def judge_on_num(
    model: StandaloneModel, mlb: Mlb, rawdata: RawData, thresh=None, n=None
):
    x = rawdata.x_test_dict
    y = rawdata.y_test_tags
    tag_num_map = defaultdict(list)
    for idx, tag in enumerate(y):
        tag_num_map[len(tag)].append(idx)

    tag_nums = []
    sizes = []
    precisions = []
    recalls = []
    f1s = []

    for tag_num, indexes in tag_num_map.items():
        num_x = [x[idx] for idx in indexes]
        num_y = [y[idx] for idx in indexes]
        if n is not None:
            pred_prob = model.over_predict(num_x, n=n)
        else:
            pred_prob = model.predict(num_x)
        preds = label_output(pred_prob, thresh)
        y_vector = mlb.transform(num_y)
        precision, recall, f1, _ = metrics.precision_recall_fscore_support(
            y_vector, preds, average="micro"
        )
        tag_nums.append(tag_num)
        sizes.append(len(indexes))
        precisions.append(precision)
        recalls.append(recall)
        f1s.append(f1)

    df = pd.DataFrame(
        {
            "Tag Number": tag_nums,
            "Count": sizes,
            "F1 Score": f1s,
            "Recall": recalls,
            "Precision": precisions,
        }
    )
    df.index = df["Tag Number"]
    return df


def judge_on_summary(summary_df: pd.DataFrame):
    def gb(df, c):
        df_ = df.copy()
        tmp_c = c + "_"
        df_[tmp_c] = df[c]
        return df_.groupby(tmp_c).sum()

    tc = gb(summary_df, "Tag Number")
    tc["Recall"] = tc["Correct Count"] / tc["Tag Number"]
    tc["Precision"] = tc["Correct Count"] / tc["Pred Tag Number"]
    tc["F1 Score"] = (
        2 * tc["Recall"] * tc["Precision"] / (tc["Recall"] + tc["Precision"])
    )
    tc["Count"] = tc["Tag Number"] / tc.index
    return tc



import random

import torch
from sklearn.preprocessing import MultiLabelBinarizer
from torch.utils.data import Dataset
from transformers import AutoTokenizer

from .data_utils import compose, upsampling
from .domain import Params

random.seed(42)


class CustomDataset(Dataset):
    def __init__(self, texts, tokenizer: AutoTokenizer, max_len: int, tags=None):
        self.tokenizer = tokenizer
        self.texts = texts
        self.tags = tags
        self.max_len = max_len

    def __len__(self):
        return len(self.texts)

    def __getitem__(self, index):
        text = self.texts[index]

        inputs = self.tokenizer.encode_plus(
            text,
            None,
            add_special_tokens=True,
            truncation=True,
            max_length=self.max_len,
            padding="max_length",
            return_token_type_ids=True,
        )
        ids = inputs["input_ids"]
        mask = inputs["attention_mask"]
        token_type_ids = inputs["token_type_ids"]

        if self.tags is None:
            return {
                "input_ids": torch.tensor(ids, dtype=torch.long),
                "attention_mask": torch.tensor(mask, dtype=torch.long),
                "token_type_ids": torch.tensor(token_type_ids, dtype=torch.long),
            }

        return {
            "input_ids": torch.tensor(ids, dtype=torch.long),
            "attention_mask": torch.tensor(mask, dtype=torch.long),
            "token_type_ids": torch.tensor(token_type_ids, dtype=torch.long),
            "labels": torch.tensor(self.tags[index], dtype=torch.float),
        }


class DatasetFactory:
    def __init__(self, params: Params):
        self.params = params
        raw_data = params.raw_data
        self.x_test_dict = raw_data.x_test_dict
        self.x_train_dict = raw_data.x_train_dict
        self.y_train_tags = raw_data.y_train_tags
        self.y_test_tags = raw_data.y_test_tags
        if params.mlb is not None:
            self.mlb = params.mlb
            self.num_labels = len(self.mlb.classes_)
        else:
            self.init_mlb()
        self.init_tokenizer(params.identifier)

    def init_mlb(self):
        y = self.y_train_tags + self.y_test_tags
        mlb = MultiLabelBinarizer().fit(y)
        self.mlb = mlb
        self.num_labels = len(self.mlb.classes_)

    def init_tokenizer(self, identifier: str):
        self.tokenizer = AutoTokenizer.from_pretrained(identifier)

    def supply_training_dataset(self):
        x_train, y_train_raw = self._upsampling(self.x_train_dict, self.y_train_tags)
        y_train = self.mlb.transform(y_train_raw)
        x_test = list(map(compose, self.x_test_dict))
        y_test = self.mlb.transform(self.y_test_tags)
        train_dataset = CustomDataset(
            x_train, self.tokenizer, self.params.max_len, y_train
        )
        testing_set = CustomDataset(x_test, self.tokenizer, self.params.max_len, y_test)
        return train_dataset, testing_set

    def _upsampling(self, x: list, y: list):
        return upsampling(
            x, y, keep_key=self.params.keep_key, target=self.params.upsampling
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tagc']

package_data = \
{'': ['*']}

install_requires = \
['dash>=1.16.2,<2.0.0',
 'fire>=0.3.1,<0.4.0',
 'pandas>=1.1.3,<2.0.0',
 'plotly>=4.11.0,<5.0.0',
 'requests>=2.24.0,<3.0.0',
 'scikit-learn>=0.23.2,<0.24.0',
 'torch>=1.6.0,<2.0.0',
 'tqdm>=4.50.1,<5.0.0',
 'transformers==3.5.1',
 'wheel>=0.35.1,<0.36.0']

setup_kwargs = {
    'name': 'tagc',
    'version': '0.1.8',
    'description': '',
    'long_description': None,
    'author': 'Kiyoshi Mu',
    'author_email': 'mooyewtsing@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.9,<4.0.0',
}


setup(**setup_kwargs)



import random
from itertools import accumulate
from typing import List

import pandas as pd

from tagc.data_utils import (
    count_tags,
    labelled_cases_to_xy,
    load_labelled_cases,
    split_and_dump_dataset,
    train_test_split,
)
from tagc.domain import RawData
from tagc.io_utils import dump_datazip, load_datazip, load_json
from tagc.review import review_pipe

random.seed(42)


def clean(x_dict, y_tags):
    content = sorted(
        (("".join(v.values()), idx) for idx, v in enumerate(x_dict)),
        key=lambda x: x[0],
    )
    for c1, c2 in zip(content, content[1:]):  # check duplication
        if c1[0] == c2[0]:
            print(f"remove duplication, idx:{c1[1]}")
            del_idx = c1[1]
            x_dict.pop(del_idx)
            y_tags.pop(del_idx)
    for idx, tag in enumerate(y_tags):
        if "normal" in tag and "iron deficiency" not in tag:
            y_tags[idx] = ["normal"]
    x = x_dict
    y = y_tags
    x_train_dict, x_test_dict, y_train_tags, y_test_tags = train_test_split(
        x, y, test_size=0.2, train_first=True
    )
    rd = RawData(x, y, x_train_dict, y_train_tags, x_test_dict, y_test_tags)
    return rd


def replay_ac(ac_data_ps: List[str], dst="."):
    """Replay the active learning data sampling results

    Args:
        ac_data_ps (List[str]): File paths of labels

    Returns:
        history (List[Counter]): label count for each iteration
        sizes (List[int]): case number for each iteration
        dsps (List[str]): File paths of datasets
    """
    history = []
    sizes = []
    dsps = []
    start = load_labelled_cases(ac_data_ps[0])
    ds = clean(*labelled_cases_to_xy(start))
    dsp = dump_datazip(ds, f"{dst}/data0.zip")
    sizes.append(len(ds.x_dict))
    history.append(count_tags(ds.y_tags).keys())
    for idx, target in enumerate(ac_data_ps[1:], 1):
        ds = clean(*review_pipe(dsp, target, return_xy=True))
        dsp = dump_datazip(ds, f"{dst}/data{idx}.zip")
        dsps.append(dsp)
        sizes.append(len(ds.x_dict))
        history.append(count_tags(ds.y_tags).keys())
    return history, sizes, dsps


def make_history_df(history, sizes):
    diffs = []
    tag_count = []
    for i in range(len(history) - 1, 0, -1):
        diff = sorted(set(history[i]) - set(history[i - 1]))
        diffs.append(diff)
        tag_count.append(len(diff))
    diff = sorted(history[0])
    diffs.append(diff)
    tag_count.append(len(diff))
    tag_count = accumulate(reversed(tag_count))
    hist_df = pd.DataFrame(
        {
            "Iteration": list(range(1, len(sizes) + 1)),
            "New labels": [", ".join(item) for item in reversed(diffs)],
            "Label Count": list(tag_count),
            "Sample Count": sizes,
        }
    )
    return hist_df


def dataset_split(final_dsp, dst="."):
    ds = load_datazip(final_dsp)
    standard_dsps = []
    for i in range(3):
        output = f"{dst}/standardDs{i}.zip"
        split_and_dump_dataset(ds.x_dict, ds.y_tags, train_first=False, output=output)
        standard_dsps.append(output)
    return standard_dsps


def form_random_ds(
    standard_dsps: List[str],
    eval_ret="mona_j.csv",
    unlabelled_p="unlabelled.json",
    outdir=".",
    train_size=400,
):
    df = pd.read_csv(eval_ret).drop_duplicates(subset=["ID", "Judge"], keep="last")
    indices = df["ID"].to_list()
    sampled_cases = load_json(unlabelled_p)
    add_text = [sampled_cases[idx] for idx in indices]
    y_true_ = df["eval"].map(lambda x: x.split(", ")).to_list()
    dsps = []
    for i, base_path in enumerate(standard_dsps):
        ds = load_datazip(base_path)
        random_idx = random.sample(list(range(len(indices))), train_size)
        x_train_dict = [add_text[idx] for idx in random_idx]
        y_train_tags = [y_true_[idx] for idx in random_idx]
        ds.x_train_dict = x_train_dict
        ds.y_train_tags = y_train_tags
        ds.x_dict = x_train_dict + ds.x_test_dict
        ds.y_tags = y_train_tags + ds.y_test_tags

        dsp = dump_datazip(ds, f"{outdir}/randomDs{i}.zip")
        dsps.append(dsp)
        print(dsp)
    return dsps



from os.path import join
import pandas as pd
from sklearn import metrics
from sklearn.preprocessing import MultiLabelBinarizer
from typing import List
from tagc.io_utils import load_json


def continue_eval(sheet_csv, y_pred_: List[list] = None):
    p_df = pd.read_csv(sheet_csv).drop_duplicates(subset=["ID", "Judge"], keep="last")
    mlb = MultiLabelBinarizer()
    # mlb.fit(p_df["eval"].map(lambda x: x.split(", ")).tolist())
    mlb.fit(y_pred_)
    p_df = p_df.groupby("Judge").agg(list)
    out = {}
    all_evals = []

    for judge, df in p_df.iterrows():

        y_true = list(map(lambda x: x.split(", "), df["eval"]))
        ids = df["ID"]
        if y_pred_ is None:
            y_pred = list(map(lambda x: x.split(", "), df["pred"]))
        else:
            y_pred = [y_pred_[id_] for id_ in ids]

        y_true_trans = mlb.transform(y_true)
        y_pred_trans = mlb.transform(y_pred)
        precision, recall, f1, _ = metrics.precision_recall_fscore_support(
            y_true_trans, y_pred_trans, average="micro"
        )

        all_evals.append(y_true_trans)
        out[judge] = {"precision": precision, "recall": recall, "f1": f1}
    return out


def form_pred(eval_json):
    eval = load_json(eval_json)
    y_pred = []
    for item in eval:
        y_pred.append(
            [prob[0] for pred, prob in zip(item["pred"], item["prob"]) if pred]
        )
    return y_pred


if __name__ == "__main__":
    base = "E:/Coding/scholar/continue"
    sheet_csv = join(".", "prediction judgement - Sheet1.csv")
    print(
        continue_eval(sheet_csv, form_pred("E:\\outputsS\\outputsS\\eval.json")),
    )



import numpy as np
from numpy import argmax
from sklearn.metrics import precision_recall_curve

from tagc.io_utils import dump_json


def cal_f1(precision, recall):
    a = 2 * precision * recall
    b = precision + recall
    fscore = np.divide(a, b, out=np.zeros_like(a), where=b != 0)
    return fscore


def analysis_kf(rets, mlb, out_folder=""):
    labels = []

    for k_fold in rets:
        k_probs = k_fold["prob"]
        labels.extend(k_fold["tag"])

    labels = mlb.transform(labels)
    probs = np.zeros_like(labels, dtype=np.float)

    thresh = {}

    for tag_idx, tag in enumerate(mlb.classes_):
        tag_probs = []
        for k_fold in rets:
            k_probs = k_fold["prob"]
            tag_probs.extend(float(prob[tag_idx][1]) for prob in k_probs)
        tag_labels = labels[:, tag_idx]
        tag_probs = np.array(tag_probs)

        probs[:, tag_idx] = tag_probs

        precision, recall, thresholds = precision_recall_curve(tag_labels, tag_probs)
        fscore = cal_f1(precision, recall)
        # locate the index of the largest f score
        ix = argmax(fscore)
        print(
            "{}: Best Threshold={:f}, F-Score={:.3f}".format(
                tag, thresholds[ix], fscore[ix]
            )
        )
        thresh[tag_idx] = thresholds[ix]

    thresh_json = []
    for k, v in thresh.items():
        thresh_json.append((mlb.classes_[k], str(v)))
    dump_json(out_folder + "thresh_info.json", thresh_json)

    thresh_v = [v if v < 0.5 else 0.5 for v in thresh.values()]
    dump_json(out_folder + "thresh.json", thresh_v)
    # return thresh_json, thresh_v
